#ifndef DALI_INTERNAL_SCENE_GRAPH_PROPERTY_OWNER_H
#define DALI_INTERNAL_SCENE_GRAPH_PROPERTY_OWNER_H

/*
 * Copyright (c) 2023 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// INTERNAL INCLUDES
#include <dali/devel-api/common/owner-container.h>
#include <dali/internal/common/const-string.h>
#include <dali/internal/common/message.h>
#include <dali/internal/update/animation/scene-graph-constraint-declarations.h>
#include <dali/internal/update/common/property-base.h>
#include <dali/internal/update/common/scene-graph-buffers.h>
#include <dali/internal/update/common/uniform-map.h>
#include <dali/public-api/common/dali-vector.h>

namespace Dali
{
namespace Internal
{
namespace SceneGraph
{
class PropertyOwner;
class ResetterManager;

using OwnedPropertyContainer = OwnerContainer<PropertyBase*>;
using OwnedPropertyIter      = OwnedPropertyContainer::Iterator;

/**
 * An update-thread object which own properties.
 * This allows observers to track the lifetime of the object & its properties.
 */
class PropertyOwner
{
public:
  class Observer
  {
  public:
    /**
     * Called when the observable object is connected to the scene graph.
     * @param[in] owner A reference to the connected PropertyOwner
     */
    virtual void PropertyOwnerConnected(PropertyOwner& owner) = 0;

    /**
     * Called when the observable object is disconnected from the scene graph.
     * @param[in] currentBufferIndex The buffer to reset.
     * @param[in] owner A reference to the disconnected PropertyOwner
     */
    virtual void PropertyOwnerDisconnected(BufferIndex updateBufferIndex, PropertyOwner& owner) = 0;

    /**
     * Called shortly before the observable object is destroyed.
     *
     * @note Cleanup should be done in both this and PropertyOwnerDisconnected as PropertyOwnerDisconnected
     * may not be called (i.e. when shutting down).
     */
    virtual void PropertyOwnerDestroyed(PropertyOwner& owner) = 0;

  protected:
    /**
     * Virtual destructor, no deletion through this interface
     */
    virtual ~Observer() = default;
  };

  /**
   * Create a property owner.
   * @return A newly allocated object.
   */
  static PropertyOwner* New();

  /**
   * Virtual destructor; this is intended as a base class.
   */
  virtual ~PropertyOwner();

  /**
   * Add an observer.
   * The observer is responsible for calling RemoveObserver(*this) during its own destruction.
   * Connecting an actor-side object as an observer is not allowed, due to thread-safety issues.
   * @param[in] observer The observer.
   */
  void AddObserver(Observer& observer);

  /**
   * Remove an observer.
   * @param[in] observer The observer.
   */
  void RemoveObserver(Observer& observer);

  /**
   * This method can be used to determine if there is an animation or
   * constraint that is using this property owner.
   * @return true if there are observers.
   */
  bool IsObserved();

  /**
   * Called just before destruction to disconnect all observers and remove constraints.
   * This occurs when the object is in the process of being destroyed.
   */
  void Destroy();

  /**
   * Notify all observers that the object has been connected
   * This occurs when the object is connected to the scene-graph during UpdateManager::Update().
   */
  void ConnectToSceneGraph();

  /**
   * Notify all observers that the object has been disconnected and remove constraints.
   * This occurs when the object is disconnected from the scene-graph during UpdateManager::Update().
   * @param[in] currentBufferIndex The current update buffer.
   */
  void DisconnectFromSceneGraph(BufferIndex updateBufferIndex);

  /**
   * Reserve the given number of properties
   */
  void ReserveProperties(int propertyCount);

  /**
   * Install a custom property.
   * @post The PropertyOwner takes ownership of the property.
   * @param[in] property A pointer to a newly allocated property.
   */
  void InstallCustomProperty(OwnerPointer<PropertyBase>& property);

  /**
   * Retrieve the custom properties owned by the object.
   * @return A container of properties.
   */
  OwnedPropertyContainer& GetCustomProperties()
  {
    return mCustomProperties;
  }

  /**
   * Retrieve the custom properties owned by the object.
   * @return A container of properties.
   */
  const OwnedPropertyContainer& GetCustomProperties() const
  {
    return mCustomProperties;
  }

  /**
   * Mark an property owner with the updated flag.
   * @param[in] updated The updated flag
   */
  void SetUpdated(bool updated)
  {
    mUpdated = updated;
  }

  /**
   * Retrieve if the property owner is updated due to the property is being animating.
   * @return An updated flag
   */
  bool Updated() const
  {
    return mUpdated;
  }

  // Constraints

  /**
   * Apply a constraint.
   * @param[in] constraint The constraint to apply.
   */
  void ApplyConstraint(OwnerPointer<ConstraintBase>& constraint);

  /**
   * Begin removal of constraints.
   * @param[in] constraint The constraint to remove.
   */
  void RemoveConstraint(ConstraintBase* constraint);

  /**
   * Retrieve the constraints that are currently applied.
   * @return A container of constraints.
   */
  ConstraintOwnerContainer& GetConstraints();

  /**
   * @copydoc UniformMap::Add
   */
  virtual void AddUniformMapping(const UniformPropertyMapping& map);

  /**
   * @copydoc UniformMap::Remove
   */
  virtual void RemoveUniformMapping(const ConstString& uniformName);

  /**
   * Get the mappings table
   */
  const UniformMap& GetUniformMap() const;

  /**
   * Query whether playing an animation is possible or not.
   * @return true if playing an animation is possible.
   */
  virtual bool IsAnimationPossible() const
  {
    return true;
  }

  /**
   * @brief Install custom resetter messages to resetter manager.
   * @pre ConnectToSceneGraph() Should be called before this API.
   *
   * @param[in] manager ResetterManager to add resetter.
   */
  virtual void AddInitializeResetter(ResetterManager& manager) const
  {
    // Do nothing
  }

protected:
  /**
   * Protected constructor.
   */
  PropertyOwner();

  /**
   * Method to inform derived classes when property maps have been modified.
   */
  virtual void OnMappingChanged()
  {
    // Default behaviour is to do nothing
  }

private:
  // Undefined
  PropertyOwner(const PropertyOwner&);

  // Undefined
  PropertyOwner& operator=(const PropertyOwner& rhs);

protected:
  OwnedPropertyContainer mCustomProperties; ///< Properties provided with InstallCustomProperty()
  UniformMap             mUniformMaps;      ///< Container of owned uniform maps
  bool                   mUpdated;
  bool                   mIsConnectedToSceneGraph;

private:
  using ObserverContainer = Dali::Vector<PropertyOwner::Observer*>;
  using ObserverIter      = ObserverContainer::Iterator;
  using ConstObserverIter = ObserverContainer::ConstIterator;

  ObserverContainer mObservers; ///< Container of observer raw-pointers (not owned)

  ConstraintOwnerContainer mConstraints; ///< Container of owned constraints
};

} // namespace SceneGraph

} // namespace Internal

} // namespace Dali

#endif // DALI_INTERNAL_SCENE_GRAPH_PROPERTY_OWNER_H
